'''
Application settings for graphite-opentsdb.
'''

from django.conf import settings

from multiprocessing.pool import ThreadPool

#: URI to the OpenTSDB API
OPENTSDB_URI = getattr(
    settings,
    'OPENTSDB_URI',
    'http://localhost:4242/api/v1/',
)

#: Tree ID to display in graphite
OPENTSDB_TREE = getattr(
    settings,
    'OPENTSDB_TREE',
    1,
)

#: Cache time for OpenTSDB fetches.
OPENTSDB_CACHE_TIME = getattr(
    settings,
    'OPENTSDB_CACHE_TIME',
    60*15,
)

#: How many concurrent requests to allow.
OPENTSDB_MAX_REQUESTS = getattr(
    settings,
    'OPENTSDB_MAX_REQUESTS',
    10,
)

#: If we know we're trying to do more than this many requests for one metric,
#: we should just do one query
OPENTSDB_METRIC_QUERY_LIMIT = getattr(
    settings,
    'OPENTSDB_METRIC_QUERY_LIMIT',
    OPENTSDB_MAX_REQUESTS,
)

#: The pool to use for concurrent requests.
#: Overrides OPENTSDB_MAX_REQUESTS if set.
OPENTSDB_REQUEST_POOL = getattr(
    settings,
    'OPENTSDB_REQUEST_POOL',
    ThreadPool(OPENTSDB_MAX_REQUESTS),
)

#: Default aggregation interval for Graphite (in seconds)
OPENTSDB_DEFAULT_AGGREGATION_INTERVAL = getattr(
    settings,
    'OPENTSDB_DEFAULT_AGGREGATION_INTERVAL',
    15,
)



from __future__ import division

from cacheback.decorators import cacheback
from django.conf import settings
from graphite.intervals import Interval, IntervalSet
from graphite.node import BranchNode, LeafNode
from graphite.readers import FetchInProgress
import re
import requests
import time
import threading

from . import app_settings

import logging
LOGGER = logging.getLogger(__name__)


class OpenTSDBNodeMixin(object):
    def __init__(self, name, *args):
        super(OpenTSDBNodeMixin, self).__init__(*args)
        self.name = name


class OpenTSDBLeafNode(OpenTSDBNodeMixin, LeafNode):
    pass


class OpenTSDBBranchNode(OpenTSDBNodeMixin, BranchNode):
    pass


def find_nodes_from_pattern(opentsdb_uri, opentsdb_tree, pattern):
    query_parts = []
    for part in pattern.split('.'):
        part = part.replace('*', '.*')
        part = re.sub(
            r'{([^{]*)}',
            lambda x: "(%s)" % x.groups()[0].replace(',', '|'),
            part,
        )
        query_parts.append(part)

    shared_reader = SharedReader()
    nodes = list(find_opentsdb_nodes(opentsdb_uri, query_parts, "%04X" % opentsdb_tree, shared_reader=shared_reader))
    shared_reader.node_count = len(nodes)
    for node in nodes:
        yield node


@cacheback(app_settings.OPENTSDB_CACHE_TIME)
def get_opentsdb_url(opentsdb_uri, url):
    full_url = "%s/%s" % (opentsdb_uri, url)
    return requests.get(full_url).json()


def find_opentsdb_nodes(opentsdb_uri, query_parts, current_branch, shared_reader, path=''):
    query_regex = re.compile(query_parts[0])
    for node, node_data in get_branch_nodes(opentsdb_uri, current_branch, shared_reader, path):
        node_name = node_data['displayName']
        dot_count = node_name.count('.')

        if dot_count:
            node_query_regex = re.compile(r'\.'.join(query_parts[:dot_count+1]))
        else:
            node_query_regex = query_regex

        if node_query_regex.match(node_name):
            if len(query_parts) == 1:
                yield node
            elif not node.is_leaf:
                # We might need to split into two branches here
                # if using dotted nodes, as we can't tell if the UI
                # wanted all nodes with a single * (from advanced mode)
                # or if a node like a.b is supposed to be matched by *.*
                if query_parts[dot_count+1:]:
                    for inner_node in find_opentsdb_nodes(
                        opentsdb_uri,
                        query_parts[dot_count+1:],
                        node_data['branchId'],
                        shared_reader,
                        node.path,
                    ):
                        yield inner_node
                if dot_count and query_parts[0] == '.*':
                    for inner_node in find_opentsdb_nodes(
                        opentsdb_uri,
                        query_parts[1:],
                        node_data['branchId'],
                        shared_reader,
                        node.path,
                    ):
                        yield inner_node


def get_branch_nodes(opentsdb_uri, current_branch, shared_reader, path):
    results = get_opentsdb_url(opentsdb_uri, "tree/branch?branch=%s" % current_branch)
    if results:
        if path:
            path += '.'
        if results['branches']:
            for branch in results['branches']:
                yield OpenTSDBBranchNode(branch['displayName'], path + branch['displayName']), branch
        if results['leaves']:
            for leaf in results['leaves']:
                reader = OpenTSDBReader(
                    opentsdb_uri,
                    leaf,
                    shared_reader,
                )
                yield OpenTSDBLeafNode(leaf['displayName'], path + leaf['displayName'], reader), leaf


class OpenTSDBFinder(object):
    def __init__(self, opentsdb_uri=None, opentsdb_tree=None):
        self.opentsdb_uri = (opentsdb_uri or app_settings.OPENTSDB_URI).rstrip('/')
        self.opentsdb_tree = opentsdb_tree or app_settings.OPENTSDB_TREE

    def find_nodes(self, query):
        for node in find_nodes_from_pattern(self.opentsdb_uri, self.opentsdb_tree, query.pattern):
            yield node


class SharedReader(object):
    def __init__(self):
        self.worker = threading.Semaphore(1)
        self.config_lock = threading.Lock()
        self.workers = {}
        self.results = {}
        self.result_events = {}

    def get(self, opentsdb_uri, aggregation_interval, leaf_data, start, end):
        key = (opentsdb_uri, aggregation_interval, leaf_data['metric'], start, end)
        with self.config_lock:
            if key not in self.workers:
                self.workers[key] = threading.Semaphore(1)
                self.result_events[key] = threading.Event()

        if self.workers[key].acquire(False):
            # we are the worker, do the work
            data = requests.get("%s/query?m=sum:%ds-avg:%s{%s}&start=%d&end=%d&show_tsuids=true" % (
                opentsdb_uri,
                aggregation_interval,
                leaf_data['metric'],
                ','.join(["%s=*" % t for t in leaf_data['tags']]),
                start,
                end,
            )).json()

            self.results[key] = {}
            for metric in data:
                assert len(metric['tsuids']) == 1
                self.results[key][metric['tsuids'][0]] = [metric]
            self.result_events[key].set()

        self.result_events[key].wait()
        tsuid = leaf_data['tsuid']
        if tsuid in self.results[key]:
            return self.results[key][tsuid]
        else:
            return []

class OpenTSDBReader(object):
    __slots__ = ('opentsdb_uri', 'leaf_data', 'shared_reader',)
    supported = True
    step = app_settings.OPENTSDB_DEFAULT_AGGREGATION_INTERVAL

    def __init__(self, opentsdb_uri, leaf_data, shared_reader):
        self.opentsdb_uri = opentsdb_uri
        self.leaf_data = leaf_data
        self.shared_reader = shared_reader

    def get_intervals(self):
        return IntervalSet([Interval(0, time.time())])

    def fetch(self, startTime, endTime):
        def get_data():

            if self.shared_reader.node_count > app_settings.OPENTSDB_METRIC_QUERY_LIMIT:
                data = self.shared_reader.get(
                    self.opentsdb_uri,
                    app_settings.OPENTSDB_DEFAULT_AGGREGATION_INTERVAL,
                    self.leaf_data,
                    int(startTime),
                    int(endTime),
                )
            else:
                data = requests.get("%s/query?tsuid=sum:%ds-avg:%s&start=%d&end=%d" % (
                    self.opentsdb_uri,
                    app_settings.OPENTSDB_DEFAULT_AGGREGATION_INTERVAL,
                    self.leaf_data['tsuid'],
                    int(startTime),
                    int(endTime),
                )).json()

            time_info = (startTime, endTime, self.step)
            number_points = int((endTime-startTime)//self.step)
            datapoints = [None for i in range(number_points)]

            for series in data:
                for timestamp, value in series['dps'].items():
                    timestamp = int(timestamp)
                    interval = timestamp - (timestamp % app_settings.OPENTSDB_DEFAULT_AGGREGATION_INTERVAL)
                    index = (interval - int(startTime)) // self.step
                    datapoints[index] = value

            return (time_info, datapoints)

        job = app_settings.OPENTSDB_REQUEST_POOL.apply_async(get_data)

        return FetchInProgress(job.get)



'''Tests for graphite-opentsdb.'''
# pylint: disable=C0103
# pylint: disable=R0904

from django import test
from django.core.cache import cache
from httmock import all_requests, with_httmock, HTTMock
import mock

from graphite_opentsdb.finder import OpenTSDBFinder
from graphite.storage import FindQuery
from graphite_opentsdb import app_settings


@all_requests
def mocked_urls(url, request):
    return {
        ('localhost:4242', '/api/v1/tree/branch', 'branch=0001'): {
            'status_code': 200,
            'content': '''
                {
                    "branchId": "0001",
                    "branches": [
                        {
                            "branchId": "0001CFE0B4A4",
                            "branches": null,
                            "depth": 1,
                            "displayName": "branch1",
                            "leaves": null,
                            "path": {
                                "0": "ROOT",
                                "1": "branch1"
                            },
                            "treeId": 1
                        },
                        {
                            "branchId": "00013FFD49C8",
                            "branches": null,
                            "depth": 1,
                            "displayName": "branch2",
                            "leaves": null,
                            "path": {
                                "0": "ROOT",
                                "1": "branch2"
                            },
                            "treeId": 1
                        }
                    ],
                    "depth": 0,
                    "displayName": "ROOT",
                    "leaves": [
                        {
                            "displayName": "leaf",
                            "metric": "leaf",
                            "tags": {
                                "host": "localhost"
                            },
                            "tsuid": "000BC700000100047A"
                        }
                    ],
                    "path": {
                        "0": "ROOT"
                    },
                    "treeId": 1
                }
            ''',
        },
        ('localhost:4242', '/api/v1/tree/branch', 'branch=0001CFE0B4A4'): {
            'status_code': 200,
            'content': '''
                {
                    "branchId": "0001CFE0B4A4",
                    "branches": null,
                    "depth": 1,
                    "displayName": "branch1",
                    "leaves": [
                        {
                            "displayName": "leaf",
                            "metric": "branch1.leaf",
                            "tags": {
                                "host": "localhost"
                            },
                            "tsuid": "000BC700000100047B"
                        }
                    ],
                    "path": {
                        "0": "ROOT",
                        "1": "branch1"
                    },
                    "treeId": 1
                }
            ''',
        },
        ('localhost:4242', '/api/v1/tree/branch', 'branch=00013FFD49C8'): {
            'status_code': 200,
            'content': '''
                {
                    "branchId": "00013FFD49C8",
                    "branches": null,
                    "depth": 1,
                    "displayName": "branch2",
                    "leaves": [
                        {
                            "displayName": "leaf",
                            "metric": "branch2.leaf",
                            "tags": {
                                "host": "localhost"
                            },
                            "tsuid": "000BC700000100047C"
                        }
                    ],
                    "path": {
                        "0": "ROOT",
                        "1": "branch2"
                    },
                    "treeId": 1
                }
            ''',
        },
        ('localhost:4242', '/api/v1/tree/branch', 'branch=0002'): {
            'status_code': 200,
            'content': '''
                {
                    "branchId": "0002",
                    "branches": [
                        {
                            "branchId": "0002CFE0B4A4",
                            "branches": null,
                            "depth": 1,
                            "displayName": "branch.with.dots",
                            "leaves": null,
                            "path": {
                                "0": "ROOT",
                                "1": "branch.with.dots"
                            },
                            "treeId": 1
                        }
                    ],
                    "depth": 0,
                    "displayName": "ROOT",
                    "leaves": null,
                    "path": {
                        "0": "ROOT"
                    },
                    "treeId": 2
                }
            ''',
        },
        ('localhost:4242', '/api/v1/tree/branch', 'branch=0002CFE0B4A4'): {
            'status_code': 200,
            'content': '''
                {
                    "branchId": "0002CFE0B4A4",
                    "branches": null,
                    "depth": 1,
                    "displayName": "branch.with.dots",
                    "leaves": [
                        {
                            "displayName": "leaf.with.dots",
                            "metric": "branch.with.dots.leaf.with.dots",
                            "tags": {
                                "host": "localhost"
                            },
                            "tsuid": "000BC700000100047D"
                        }
                    ],
                    "path": {
                        "0": "ROOT",
                        "1": "branch.with.dots"
                    },
                    "treeId": 2
                }
            ''',
        },
        ('localhost:4242', '/api/v1/tree/branch', 'branch=0003'): {
            'status_code': 404,
            'content': '',
        },
    }.get(
        (url.netloc, url.path, url.query),
        {
            'status_code': 500,
            'content': '',
        }
    )

@all_requests
def bad_urls(url, request):
    return {
        'status_code': 500,
        'content': '',
    }


class OpenTSDBFinderTestCase(test.TestCase):
    '''Test the finder class.'''

    def setUp(self):
        #self.settings_dict = copy.deepcopy(self.BASE_SETTINGS)
        self.finder = OpenTSDBFinder('http://localhost:4242/api/v1', 1)
        cache.clear()

    @mock.patch.object(app_settings, 'OPENTSDB_URI', 'http://localhost:9999')
    @mock.patch.object(app_settings, 'OPENTSDB_TREE', 999)
    def test_finder_settings(self):
        '''
        Test that the finder can default to Django settings.
        '''

        finder = OpenTSDBFinder()
        self.assertEqual(finder.opentsdb_uri, 'http://localhost:9999')
        self.assertEqual(finder.opentsdb_tree, 999)

    @with_httmock(mocked_urls)
    def test_finder_nodes(self):
        '''
        Test that the finder can find nodes.
        '''

        # Base query
        nodes = list(self.finder.find_nodes(query=FindQuery('*', None, None)))

        self.assertEqual(
            [node.path for node in nodes],
            ['branch1', 'branch2', 'leaf'],
        )

        # One level down
        nodes = list(self.finder.find_nodes(query=FindQuery('*.leaf', None, None)))

        self.assertEqual(
            [node.path for node in nodes],
            ['branch1.leaf', 'branch2.leaf']
        )

        nodes = list(self.finder.find_nodes(query=FindQuery('branch1.leaf', None, None)))

        self.assertEqual(
            [node.path for node in nodes],
            ['branch1.leaf'],
        )

    @with_httmock(mocked_urls)
    def test_finder_braces(self):
        '''
        Test that the finder can deal with brace expressions.
        '''

        nodes = list(self.finder.find_nodes(query=FindQuery('{branch1,leaf}', None, None)))
        self.assertEqual(
            [node.path for node in nodes],
            ['branch1', 'leaf'],
        )

    @with_httmock(mocked_urls)
    def test_finder_character_classes(self):
        '''
        Test that the finder can deal with character classes.
        '''

        nodes = list(self.finder.find_nodes(query=FindQuery('branch[1]', None, None)))
        self.assertEqual(
            [node.path for node in nodes],
            ['branch1'],
        )

    @with_httmock(mocked_urls)
    def test_finder_dotted_nodes(self):
        '''
        Test that the finder can deal with nodes with dots
        (Since OpenTSDB allows that).
        '''

        finder = OpenTSDBFinder('http://localhost:4242/api/v1/', 2)

        nodes = list(finder.find_nodes(query=FindQuery('*', None, None)))
        self.assertEqual(
            [node.name for node in nodes],
            ['branch.with.dots'],
        )

    @with_httmock(mocked_urls)
    def test_finder_nested_dotted_nodes(self):
        '''
        Test that the finder can deal with nested nodes with dots
        (Since OpenTSDB allows that).
        '''

        finder = OpenTSDBFinder('http://localhost:4242/api/v1/', 2)

        nodes = list(finder.find_nodes(query=FindQuery('*.*', None, None)))
        self.assertEqual(
            [node.name for node in nodes],
            ['leaf.with.dots'],
        )

    @with_httmock(mocked_urls)
    def test_finder_missing_branch(self):
        '''
        Test that the finder can deal with a missing branch.
        '''

        finder = OpenTSDBFinder('http://localhost:4242/api/v1/', 3)

        with self.assertRaises(ValueError):
            nodes = list(finder.find_nodes(query=FindQuery('*', None, None)))


    def test_caching(self):
        '''
        Test caching behaviour.
        '''
        with HTTMock(mocked_urls):
            nodes = list(self.finder.find_nodes(query=FindQuery('*', None, None)))

            self.assertEqual(
                [node.path for node in nodes],
                ['branch1', 'branch2', 'leaf'],
            )

        with HTTMock(bad_urls):
            nodes = list(self.finder.find_nodes(query=FindQuery('*', None, None)))

            self.assertEqual(
                [node.path for node in nodes],
                ['branch1', 'branch2', 'leaf'],
            )



'''Setup file for graphite-opentsdb.'''
from setuptools import find_packages, setup
from graphite_opentsdb.version import __VERSION__

setup(
    name                 ='graphite-opentsdb',
    version              = __VERSION__,
    packages             = find_packages(),
    description          = 'A graphite storage plugin for OpenTSDB.',
    author               = 'Mike Bryant',
    author_email         = 'mike@mikebryant.me.uk',
    install_requires     = ['django', 'django-cacheback', 'graphite-web', 'requests'],
    include_package_data = True,
    test_suite           = 'setuptest.setuptest.SetupTestSuite',
    tests_require        = ['django-setuptest', 'httmock', 'mock'],
    url                  = 'https://github.com/mikebryant/graphite-opentsdb-finder',
    classifiers          = [
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
    ],
)
